"""Samplers for the Gaussian and jump parts of the Levy basis; we include the drift 
   term in the Gaussian part. The distributions are parametrised as in
           https://docs.scipy.org/doc/scipy/reference/stats.html"""
         
import numpy as np
from scipy.stats import norm,gamma,cauchy,invgauss,norminvgauss,\
                            geninvgauss,poisson             

def gaussian_part_sampler(gaussian_part_params,areas):
    """Simulates the Gaussian part (including drift) of the Levy basis over disjoint sets

    Args:
      gaussian_part_params: list or numpy array with the drift term and variance of the Gaussian part   
      areas: A number / numpy arrray containing the areas of the given sets

    Returns:
      A number / numpy array with law \(\mathcal{N}(drift \cdot areas, scale \cdot \sqrt{areas})\); we use the
      mean-scale parametrisation for consistency with scipy
    """
    
    drift,scale = gaussian_part_params
    gaussian_sample = norm.rvs(loc = drift * areas, scale = scale *(areas)**0.5)
    return gaussian_sample
    
def jump_part_sampler(jump_part_params,areas,distr_name):
    """Simulates the jump part of the Levy basis over disjoint sets; distributions are named 
    and parametrised as in https://docs.scipy.org/doc/scipy/reference/stats.html
    
    Args:
      distr_name: Name of the distribution of the jump part L_j
      jump_part_params: List or numpy array which contains the parameters of the 
      distribution of the jump part L_j
      areas: A number/numpy arrray containing the areas of the given sets
    
    Returns:
      A number / numpy array with law specified by params and distr_name
    """

    if distr_name == None:
        samples = np.zeros(shape= areas.shape)
    
    ###continuous distributions
    elif distr_name == 'gamma':
        a,scale = jump_part_params
        samples = gamma.rvs(a = a * areas, loc = 0, scale = scale)
    
    elif distr_name == 'cauchy':
        scale = jump_part_params[0]
        samples = cauchy.rvs(loc = 0, scale = scale * areas)
        
    elif distr_name == 'invgauss':
        #this is a different parametrisation
        #from the wikipedia page
        raise ValueError('not implemented')
        
    ###discrete distributions
    elif distr_name == 'poisson':
        lambda_poisson = jump_part_params[0]
        samples = poisson.rvs(mu = lambda_poisson * areas,loc=0)
    

    return samples       



import numpy as np
import pandas as pd 

#check qlike definition
def qlike_func(forecast,observed):
    return observed/forecast - np.log(observed/forecast)  - 1

#check the ones below and the forecast as well from trawl ( where we actually do the forecast)
def loss_func(trawl_values,predicted_values_dict,training_window):
    #mae - mean absolute error
    #medae - mean of the median absolute error
    #rmse - root mean squared error
    #qlike
    assert isinstance(predicted_values_dict,dict)
    assert isinstance(trawl_values,np.ndarray)
    
    assert len(trawl_values.shape) == 2
    print('here')
    mae,medae,rmse,mqlike,nr_steps_ahead = [],[],[],[],[]
    
    for h in predicted_values_dict.keys(): 
        print(predicted_values_dict.keys())
        true_values = trawl_values[:,training_window + h-1:]
        print(true_values.shape,predicted_values_dict[h].shape)
        assert true_values.shape == predicted_values_dict[h].shape
        
        #check what it does
        #mae
        mae.append( np.mean(np.abs(true_values - predicted_values_dict[h])) )
        #mean median error
        medae.append( np.mean(np.median(np.abs(true_values - predicted_values_dict[h]),axis=1)) )
        #mse
        mse_vec = np.mean((true_values - predicted_values_dict[h])**2,axis=1)
        #check
        assert len(mse_vec.shape) == 1 and  len(mse_vec) == true_values.shape[0]
        rmse.append(np.mean(mse_vec**0.5)) 
        #mean qlike loss
        mqlike.append(np.mean(qlike_func(predicted_values_dict[h],true_values))) 
        nr_steps_ahead.append(h)
    df = pd.DataFrame(np.array([mae,medae,rmse,mqlike,nr_steps_ahead]).T,\
                      columns = ['mae', 'medae','rmse','mqlike','nr_steps_ahead'])
    return df

    
   
    
    
    


###################################
#module imports
from scipy.integrate import quad
import numpy as np
import pandas as pd
import math

#from scipy.optimize import minimize
from scipy.signal import convolve2d #flip the filter 
#from statsmodels.tsa.stattools import acf
#from scipy.stats import norm,gamma,cauchy,invgauss,norminvgauss,\
#                           geninvgauss,poisson 

#helper function imports
from .helpers.input_checks  import check1
from .helpers.input_checks  import check_grid_params
#from .helpers.input_checks  import check_cpp_params
from .helpers.input_checks  import check_trawl_function
from .helpers.input_checks  import check_jump_part_and_params
from .helpers.input_checks  import check_gaussian_params

 
from .helpers.sampler       import gaussian_part_sampler 
from .helpers.sampler       import jump_part_sampler
from .helpers.alternative_convolution_implementation import cumulative_and_diagonal_sums

#from .helpers.loss_functions import qlike_func
#helper_module = import_file(os.path.join(Path().resolve().parent,'helpers','loss_functions'))

#qlike = helper_module.qlike_func
####################################

class trawl:
    def __init__(self,nr_trawls, nr_simulations, trawl_function=None, tau = None, 
                decorrelation_time = -np.inf, mesh_size = None, times_grid = None, 
                truncation_grid = None, truncation_cpp = None, gaussian_part_params=None, 
                jump_part_name=None,jump_part_params=None,cpp_part_name = None, 
                cpp_part_params = None, cpp_intensity = None, values=None):
        """A container class for the simulation, parameter inference and forecasting of trawl processes.
        
        The implemented simulation algorithms are the grid, slice and cpp algorithms, as described in [paper link].
        
        The arguments required for the `simulate` method are `nr_trawls`,`nr_simulations`,
        `trawl_function` and `gaussian_part_params`. Further, the slice method requires `tau`,
        `decorrelation_time`,`jump_part_name`,`jump_part_params`, the grid method requires
        `mesh_size`,`times_grid`,`truncation_grid`,`jump_part_name`,`jump_part_params` and
        the cpp method requires `truncation_cpp,`,`cpp_part_name`, 
        `cpp_part_params`, `cpp_intensity`.
               
        Args:
          nr_simulation: positive integer: number of simulations.
          nr_trawls: positive integer: number of trawls on the time axis.
          trawl_function: a non-negative, continuous, strictly increasing function \(\phi \colon (-\infty,0] \\to [0,\infty)\)
          with \(\phi(1) =0, \phi(t) =0\) for \(t>0\).
          gaussian_part_params: tuple with the mean and standard deviation of the Gaussian Part.
          tau: positive number: spacing between ambit sets on the time axis, used in the slice simulation method.
          decorrelation_time: \(-\infty\) if the ambit set A is unbounded and finite, negative otherwise, used in the slice simulation method.
          jump_part_name: a
          jump_part_params: b
          truncation_grid: c
         """
       
        #general attributes
        check1(nr_trawls, nr_simulations)
        self.nr_trawls                      = nr_trawls
        self.nr_simulations                 = nr_simulations
        
        #attributes required for simulation
        self.trawl_function                 = trawl_function
        self.gaussian_part_params           = gaussian_part_params

        
        #############################################################################     
        ### attributes required only for the slice partition simulation algorithm ###  
        
        self.tau                            = tau
        self.decorrelation_time             = decorrelation_time
        self.I                              = None
        self.slice_areas_matrix             = None
        # distributional parameters of the gaussian and jump parts of the levy seed 
        # jump_part_name and jump_part_params are also requried for the grid method
        self.jump_part_name                 = jump_part_name
        self.jump_part_params               = jump_part_params
        
        ################################################################## 
        ### attributes required only for the grid simulation algorithm ###
        
        self.times_grid               = times_grid
        self.truncation_grid          = truncation_grid       
        self.mesh_size                = mesh_size
        self.times_grid               = times_grid
        self.vol                      = None
        
        #self.indicator_matrix               = None
        
        ################################################################## 
        ### attributes required only for the cpp simulation algorithm  ###
        
        self.cpp_part_name   = cpp_part_name 
        self.cpp_part_params = cpp_part_params
        self.cpp_intensity   = cpp_intensity
            
        ### arrays containing the gaussian, jump and cpp parts of the simulation
        self.gaussian_values          =   np.empty(shape = [self.nr_simulations,self.nr_trawls])
        self.jump_values              =   np.empty(shape = [self.nr_simulations,self.nr_trawls])
        self.cpp_values               =   np.empty(shape = [self.nr_simulations,self.nr_trawls])
        
        ### passed by the user or simulated using one of the simulation methods ### 
        self.values                   =   values 
        
        
        ##################################################################
        ### attributes required only for the cpp simulation algorithm  ###
        
        self.cpp_part_name                  = cpp_part_name
        self.cpp_part_params                = cpp_part_params
        self.cpp_intensity                  = cpp_intensity
        
    ######################################################################
    ###  Simulation algorithms:      I slice, II grid, III cpp         ###
    ######################################################################
    
    
       ########################### I Slice ###########################

    
    def compute_slice_areas_finite_decorrelation_time(self):
        """Computes the \(I \\times k\) matrix 
        
        
        corresponding to the areas of the slices 
        
                
        \[\\begin{bmatrix}
        L(S_{11})  & \\ldots & L(S_{1,k-1}) & L(S_{1k}) \\\\
        L(S_{21})  &  \\ldots &  L(S_{2,k-1}) & L(S_{2k}) \\\\
         \\vdots &       &  \\vdots  & \\vdots  \\\\
        L(S_{I1}) &  \\ldots &  L(S_{I,k-1}) & L(S_{I,k})
        \\end{bmatrix}
        \]"""
        self.I = math.ceil(-self.decorrelation_time/self.tau)
        
        s_i1 = [quad(self.trawl_function,a=-i *self.tau, b = (-i+1) * self.tau)[0]
                for i in range(1,self.I+1)]
        s_i2 = np.append(np.diff(s_i1[::-1])[::-1],s_i1[-1])
        
        right_column = np.tile(s_i2[:,np.newaxis],(1,self.nr_trawls-1))
        left_column  = left_column  = (np.array(s_i1))[:,np.newaxis]
        self.slice_areas_matrix  = np.concatenate([left_column,right_column],axis=1)
        
        #to add I-1 columns of length I zeros
        #check entire program here and comapre with previous versions
                
    
    def compute_slice_areas_infinite_decorrelation_time(self):
        """Computes the  \(k \\times k\) matrix  
        
        \[\\begin{bmatrix}
              a_0     &  a_0 - a_1  & a_0 - a_1  &    \\ldots  & a_0 - a_1  & a_0 - a_1  & a_0 \\\\
              a_1     &  a_1 - a_2  & a_1 - a_2  &    \\ldots  & a_1 - a_2  & a_1        & 0   \\\\
              a_2     &  a_2 - a_3  & a_2 - a_3  &    \\ldots  & a_2        & 0          & 0   \\\\
                      &             &            &    \\vdots  &            &            &     \\\\
              a_{k-2} & a_{k-1}     & 0          &    \\ldots  &  0         & 0          & 0   \\\\
              a_{k-1} & 0           & 0          &             &  0         & 0          & 0
        \\end{bmatrix}\]
            
        corresponding to the areas of the slices 
        
                
        \[\\begin{bmatrix}
        L(S_{11})  & \\ldots & L(S_{1k,-1}) & L(S_{1k}) \\\\
        L(S_{21})  &  \\ldots &  L(S_{2,k-1}) & 0 \\\\
         \\vdots &       &  \\vdots  & \\vdots  \\\\
        L(S_{k1}) &  \\ldots &  0 & 0
        \\end{bmatrix}
        \]
         
        
        where \(k =\) `self.nr_trawls` and 
        
        \[\\begin{align}
        a_0 &= \int_{-\\tau}^0 \phi(u)du, \\\\
            \\vdots & \\\\
        a_{k-2} &= \int_{(-k+1)\\tau} ^{(-k+2)  \\tau} \phi(u) du, \\\\
        a_{k-1} &= \int_{-\infty}^{(-k+1)\\tau} \phi(u)du.
        \\end{align} 
        \]
        """
        s_i1 =  [quad(self.trawl_function,a=-i *self.tau, b = (-i+1) * self.tau)[0]
                 for i in range(1,self.nr_trawls)] + [quad(self.trawl_function,a=-np.inf,
                                                           b=(-self.nr_trawls+1)*self.tau)[0]]
        
                                                           
        #  a[0] -a[1] ,a[1] -a[2], ... , a[k-2] - a[k-1] , 0
        differences   = np.append(np.diff(s_i1[::-1])[::-1],0) 

        left_column  = np.array(s_i1)[:,np.newaxis]  
        right_column = np.zeros((self.nr_trawls,1))
        #we reconstruct the elements on the secondary diagonal at the end
        
                                                     
        middle_matrix = np.tile(differences[:,np.newaxis],(1,self.nr_trawls-2))
        whole_matrix  = np.concatenate([left_column,middle_matrix,right_column],axis=1)
        whole_matrix_reversed   = np.triu(np.fliplr(whole_matrix), k=0)
        np.fill_diagonal(whole_matrix_reversed,s_i1)     

        self.slice_areas_matrix = np.fliplr(whole_matrix_reversed)
    
    def simulate_slice_finite_decorrelation_time(self,slice_convolution_type):
        """helper for the `simulate_slice` method"""
        
        filter_ = np.fliplr(np.tril(np.ones(self.I),k=0)) 
        zero_matrix  = np.zeros([self.I,self.I-1])

        for simulation_nr in range(self.nr_simulations):
            gaussian_slices = gaussian_part_sampler(self.gaussian_part_params,self.slice_areas_matrix) 
            jump_slices     = jump_part_sampler(self.jump_part_params,self.slice_areas_matrix,self.jump_part_name)
            
            if slice_convolution_type == 'fft':
        
                gaussian_slices = np.concatenate([zero_matrix,gaussian_slices],axis=1)
                jump_slices = np.concatenate([zero_matrix,jump_slices],axis=1)
        
                #matrix with 1's on and below the secondary diagonal
                #flip the filter to agree with np convention
                self.gaussian_values[simulation_nr,:] = convolve2d(gaussian_slices,filter_[::-1,::-1],'valid')[0]
                self.jump_values[simulation_nr,:]     = convolve2d(jump_slices,filter_[::-1,::-1],'valid')[0]
        
            elif slice_convolution_type == 'diagonals':
                
                self.gaussian_values[simulation_nr,:] = cumulative_and_diagonal_sums(gaussian_slices)
                self.jump_slices_cumulative_sum[simulation_nr,:] =  cumulative_and_diagonal_sums(jump_slices)
                
                
    def simulate_slice_infinite_decorrelation_time(self,slice_convolution_type):
        """Helper for the `simulate_slice` method."""

        zero_matrix     = np.zeros([self.nr_trawls,self.nr_trawls-1])
        filter_         = np.fliplr(np.tril(np.ones(self.nr_trawls),k=0)) 


        for simulation_nr in range(self.nr_simulations):
            
            gaussian_slices = gaussian_part_sampler(self.gaussian_part_params,self.slice_areas_matrix) 
            jump_slices     = jump_part_sampler(self.jump_part_params,self.slice_areas_matrix,self.jump_part_name)
            
            if slice_convolution_type == 'fft':
                gaussian_slices = np.concatenate([zero_matrix,gaussian_slices],axis=1)
                jump_slices     = np.concatenate([zero_matrix,jump_slices],axis=1)
                
                self.gaussian_values[simulation_nr,:] = convolve2d(gaussian_slices,filter_[::-1,::-1],'valid')[0]
                self.jump_values[simulation_nr,:]     = convolve2d(jump_slices,filter_[::-1,::-1],'valid')[0]
                
            elif slice_convolution_type == 'diagonals':
                
                self.gaussian_values[simulation_nr,:] = cumulative_and_diagonal_sums(gaussian_slices)
                self.jump_slices_cumulative_sum[simulation_nr,:] =  cumulative_and_diagonal_sums(jump_slices)
        
        raise ValueError('not yet implemented')    
        
    def simulate_slice(self,slice_convolution_type):
        """implements algorithm [] from []. `slice_convolution_type` can be either [to add]"""
        if self.decorrelation_time == -np.inf:
            self.compute_slice_areas_infinite_decorrelation_time()
            self.simulate_slice_infinite_decorrelation_time(slice_convolution_type)
             
        elif self.decorrelation_time > -np.inf:
            assert(self.trawl_function(self.decorrelation_time)) == 0,'please check decorrelation time' 
            self.compute_slice_areas_finite_decorrelation_time()
            self.simulate_slice_finite_decorrelation_time(slice_convolution_type)
        self.values = self.gaussian_values + self.jump_values

                      
       ############################ II Grid ############################
    
    def grid_creation(self,min_t,max_t):
        """Creates a grid on \([0,\phi(0)] \\times [\\text{min_t}, \\text{max_t}]\). Each cell is represented by 
        the coordinates of its bottom left corner. To each cell we associate a sample from each of the gaussian
        and jump parts of the trawl process.
        
        Returns:
          gaussian_values:  array with the Gaussian of the Levy basis evaluated over the cells  
          jump_values: array with the jump part of the Levy basis evaluated over the cells      
        """
            
        coords = np.mgrid[0:self.trawl_function(0):self.mesh_size,min_t:max_t:self.mesh_size]
        x, t   = coords[0].flatten(), coords[1].flatten() 
        
        areas            = self.vol * np.ones([self.nr_simulations,len(t)])
        gaussian_values  = gaussian_part_sampler(self.gaussian_part_params,areas)
        jump_values      = jump_part_sampler(self.jump_part_params,areas,self.jump_part_name)
        
            
        return x,t,gaussian_values,jump_values
    
    def grid_update(self,i,t,gaussian_values,jump_values):
        """Inputs the values of the Levy basis evaluated over the grid cells on \([\\tau_{i-1}+\\text{truncation_grid},\\tau_{i-1}] \\times [0,\\phi(0)]\),
        removes the values corresponding to cells with time coordinates less than \(\\tau_{i} + \\text{truncation_grid}\) and adds new samples
        for the levy basis evaluated over the grid cells with time coordinates in \([\\tau_{i-1},\\tau_i]\) (see figure). 
        Assumes that the consecutive ambit sets at times \(\\tau_{i-1},\\tau_i\) are not disjoint, i.e.
        \(\\tau_i + \\text{truncation_grid} < \\tau_{i-1}\).
        
        Args:
          i: index of the trawl to be simulated
          t: time coordinates of the cells of the grid on \([\\tau_{i-1},\\tau_{i-1}+\\text{truncation_grid}] \\times [0,\phi(0)]\)
          gaussian_values: gaussian values for the grid on \([\\tau_{i-1},\\tau_{i-1}+\\text{truncation_grid}] \\times [0,\phi(0)]\)
          jump_values: jump values for the grid on \([\\tau_{i-1},\\tau_{i-1}+\\text{truncation_grid}] \\times [0,\phi(0)\)
        
        Returns:
          gaussian_values: gaussian values for the grid cells on \([\\tau_{i},\\tau_{i}+\\text{truncation_grid}] \\times [0,\phi(0)]\)
          jump_values: jump values for the grid cells on \([\\tau_{i},\\tau_{i}+\\text{truncation_grid}] \\times [0,\phi(0)]\)                                                                                                                                                                                                
        """
          

        ind_to_keep       =  t >= (self.times_grid[i] + self.truncation_grid)
        t[~ind_to_keep]  += -self.truncation_grid
        
        areas             = self.vol * np.ones([self.nr_simulations,sum(~ind_to_keep)])
        #print(gaussian_values[:,~ind_to_keep].shape)
        #print(self.gaussian_part_sampler(areas).shape)
        gaussian_values[:,~ind_to_keep] = gaussian_part_sampler(self.gaussian_part_params,areas)
        jump_values[:,~ind_to_keep]     = jump_part_sampler(self.jump_part_params,areas,self.jump_part_name)  
        #print('ind to keep sum is ', ind_to_keep.sum())
        #print('gaussian is ',gaussian_values.shape)
        #print('non_gaussian_values ',non_gaussian_values[:,ind_to_keep].shape)
        #print('new_gaussian_values ',new_gaussian_values.shape)
        #print('t new shape is ',t.shape)
        #print('x shape is', x.shape)
        
        return t,gaussian_values,jump_values   
    
    
    def simulate_grid(self):
        """Simulate the trawl proces at times `self.times_grid`, which don't have to be
        equally distant, via the grid method."""
        for i in range(len(self.times_grid)):
             if (i==0) or (self.times_grid[i-1] <= self.times_grid[i] + self.truncation_grid):
                 #check that we are creating the grid for the first time or that 
                 #trawls at time i-1 and i have empty intersection
                 x,t,gaussian_values, jump_values = self.grid_creation(self.times_grid[i] + self.truncation_grid, self.times_grid[i])

                
             elif self.times_grid[i-1] > self.times_grid[i] + self.truncation_grid:
                 #check that we have non empty intersection and update the grid
                 t,gaussian_values,jump_values = self.grid_update(i,x,t,gaussian_values,jump_values)
            
             indicators = x < self.trawl_function(t-self.times_grid[i])
             #print(gaussian_values.shape,indicators.shape)
             self.gaussian_values[:,i]  = gaussian_values @ indicators
             self.jump_values[:,i]      = jump_values @ indicators
             
        self.values = self.gaussian_values + self.jump_values
       ########################### III cpp ###########################
       
    def simulate_cpp():
        """ text to be added"""
        pass

       ####################### simulate meta-method #######################
    def simulate(self,method,slice_convolution_type='diagonals'):
         """Function to simulate from the trawl function. Contains sanity checks
         for the simulation parameters and uses helper functions for each simulation
         method.
         
         Args:
           method: one of the strings cpp, grid or slice
           slice_convolution_type: if method is set to slice, this can be one of 
           the strings diagonals or ftt, depending on the way we add up the simulated
           slice. This argument is ignored if method is set to grid or cpp."""
        
         check_trawl_function(self.trawl_function)
         check_gaussian_params(self.gaussian_part_params)
         
         assert method in {'cpp','grid','slice'},'simulation method not supported'
         
         if method == 'grid':
             check_jump_part_and_params(self.jump_part_name,self.jump_part_params)
             check_grid_params(self.mesh_size,self.truncation_grid,self.times_grid)
             self.vol = self.mesh_size **2
             self.simulate_grid()
             
         elif method == 'cpp':
             self.simulate_cpp()
                 
         elif method == 'slice':
             assert slice_convolution_type in {'fft','diagonals'}
             check_jump_part_and_params(self.jump_part_name,self.jump_part_params)

             self.simulate_slice(slice_convolution_type)         
             
    def theoretical_acf(self,t_values):
        """Computes the theoretical acf of the trawl process
        
        Args:
          t_values: array of time values 
        
        Returns:
          d_acf: a dictionary of the type  \(t: \\text{corr}(X_0,X_t)\), where \(t\) ranges over the input array `t_values`. 
        """
        total_area = quad(self.trawl_function,a=-np.inf,b= 0)[0]
        d_acf=dict()
        for t in t_values:
            d_acf[t] = quad(self.trawl_function,a=-np.inf,b= -t)[0]/total_area
        return d_acf
                  

             



###################################################################
#imports 
from collections     import Counter
from itertools       import chain
from scipy.optimize  import fsolve
from scipy.integrate import quad
import numpy as np
import math
import time
#from numba import njit

from .helpers.input_checks import check_trawl_function
from .helpers.input_checks  import check_jump_part_and_params
from .helpers.input_checks  import check_gaussian_params
from .helpers.input_checks import check_spatio_temporal_positions

from .helpers.sampler import gaussian_part_sampler
from .helpers.sampler import jump_part_sampler
###################################################################

class simple_ambit_field:
    def __init__(self, x, tau, k_s, k_t, nr_simulations, ambit_function=None, decorrelation_time=-np.inf,
                 gaussian_part_params=None, jump_part_name=None, jump_part_params=None,
                 batch_size=None, total_nr_samples=None, values=None):
        """Container class for the simulation, parameter inference and forecasting of ambit fields of the
        form \(Y_t(x) = L(A+(x,t))\).
        
        Args:
          x: positive number: spacing between ambit sets on the space axis.
          tau: positive number: spacing between ambit sets on the time axis.
          k_s: positive integer: number of ambit sets on the space axix.
          k_t: positive integer: number of ambit sets on the time axis.
          nr_simulation: positive integer: number of simulations.
          ambit_function: a non-negative, continuous, strictly increasing function
          \(\phi \colon (-\infty,0] \\to [0,\infty)\) with \(\phi(1) > 0, \phi(t) =0\) for \(t>0\).
          decorrelation_time: \(-\infty\) if the ambit set A is unbounded and finite, negative otherwise.
          gaussian_part_params: tuple with the mean and standard deviation of the Gaussian part.                                                           
          jump_part_name: tuple with the parameters of the jump part distribution check `helpers.sampler`
          for the parametrisation.
          jump_part_params: string: name of the jump part distribution. check `helpers.sampler` for the parametrisation.
          batch_size: positive integer: number of points to be used at once in the `approximate_slices` method,
          in order to optimise for cache memory.
          total_nr_samples: positive integer: total number of points to be used in the `approximate_slices` method.
          values: a numpy array with shape \([\\text{nr_simulations},k_s,k_t]\) which is passed by the user or 
          simulated with the method `simple_ambit_field.simulate`.
        """

        #################################################################################
        check_spatio_temporal_positions(x, tau, k_s, k_t, nr_simulations)
        self.x = x
        self.tau = tau
        self.k_s = k_s
        self.k_t = k_t
        self.nr_simulations = nr_simulations

        ### simulation parameters ###
        self.ambit_function = ambit_function
        self.gaussian_part_params = gaussian_part_params
        self.jump_part_name = jump_part_name
        self.jump_part_params = jump_part_params
        self.decorrelation_time = decorrelation_time
        self.total_nr_samples = total_nr_samples
        self.batch_size = batch_size

        ### dimension of the indicator matrix for each minimal slice 
        ### if decorrelation_time >  -inf, I_t = ceiling(decorrelation_time/tau)
        ### if decorrelation_time =  -inf, I_t = k_t - T/tau,
        ### where T = tau * floor{\phi^{-1}(x)/tau + 1} ###
        self.I_t = None
        self.I_s = None   #I_s = math.ceil(self.ambit_function(0)/self.x)


        ### minimal slices on t > T/tau ###
        self.unique_slices = None
        self.unique_slices_areas = None
        
        ### correction slices given by the intersections of ambit sets \(A_ij\) with \(1 \le i \le k_s, 1 \le j \le k_t\)
        ### with \(t < T/tau\), which we list from left to right in an array
        self.correction_slices_areas = None
        
        ### container for gaussian and jump parts. their sum is the result ###
        self.gaussian_values = None
        self.jump_values = None
        
        ### passed by the user or simulated using the simulate method 
        ### must have shape [nr_simulations,k_s,k_t] ###
        if values == None:
            self.values = None

        else:

            assert isinstance(
                values, np.ndarray), 'the values argument is not a numpy array'
            assert values.shape == (
                nr_simulations, k_s, k_t), 'please check the shape of the values argument'
            self.values = values

        #########################################################################################
        ### infered simulation parameters: skip this if you are only interested in simulations###
        self.inferred_parameters = None

        # self.inferred_parameters is a list with elements dictionaries of the form
        # {'inferred_ambit_function_name':   , inferred_ambit_function_params:    ,
        # 'inferred_gaussian_params':        ,'inferred_jump_params':              }

        # inferred_ambit_function        is 'exponential','gamma', 'ig' or a lambda function
        # inferred_ambit_function_params is a tuple
        # inferred_gaussian_params       is a tuple containing the mean and scale
        # inferred_jump_params           is a dictionary containing the name of the distribution
        # and its params, such as {'gamma': (1,1)}
        ##########################################################################################

    def delete_values(self):
        """Deletes the `values` attribute"""
        if self.values != None:
            self.values = None
            print('self.values has been deleted')
        #else:
            #print('no values to delete')

    def determine_slices_from_points(self, points_x, points_t):
        """Helper for the 'approximate_slices' method. Given random points with coordinates 
        `points_x` and `points_t` coordinates from the uniform distribution on \([x,x+\phi(0)] \\times [0,\\tau]\)
        which do not belong to \(A_{01}\) or \(A_{10}\), we check in which slice \(S\) of \(\mathcal{S}_{kl}\) each point is.
        we do so by using a 3d array with shape  \([\\text{nr_sampled_points},I_s,I_t]\) where the \(i^{\\text{th}}\) element \([i,:,:]\) is a matrix with \(\\text{kl}^{\\text{th}}\) element
        is a boolean given by \(x_i - x \cdot k < \phi(t_i -t \cdot l) \cdot (T < t_i-t \cdot l <0)\)
        where \((x_i,t_i)\)  are the coordinates of the \(i^{\\text{th}}\) uniform random sample. 
        [check description of the indicator here]
        
        Args:\n 
          points_x: x coordinates of the uniformly sampled points
          points_t: t coordinates of the uniformly sampled points
        
        Returns:
          a dictionary with keys given by tuples which represent the indicator matrices of 
          a minimal slice and values given by the number of points contained in the minimal slice
        """

        # coordinates at which the ambit field is simulated
        ambit_t_coords = self.tau * np.arange(1, self.I_t+1)
        ambit_x_coords = self.x * np.arange(1, self.I_s+1)
        
        x_ik = np.subtract.outer(points_x, ambit_x_coords)
        x_ikl = np.repeat(x_ik[:, :, np.newaxis], repeats=self.I_t, axis=2)
        t_il = np.subtract.outer(points_t, ambit_t_coords)
        phi_t_ikl = np.repeat(self.ambit_function(
            t_il)[:, np.newaxis, :], repeats=self.I_s, axis=1)
        range_indicator = x_ik > 0
        range_indicator = np.repeat(
            range_indicator[:, :, np.newaxis], repeats=self.I_t, axis=2)

        indicator = (x_ikl < phi_t_ikl) * range_indicator
        
        #in the unlikely case no minimal slice is identified
        if len(indicator) == 0:
            raise ValueError('use more samples in each batch')

        # we enumerate the unique indicator matrices together with the frequency counts
        # we change the shape from [total_nr_samples, I_s, I_t]  to [total_nr_samples, I_s * I_t]
        reshaped_indicator = indicator.reshape(indicator.shape[:-2]+(-1,))
        g = (tuple(i) for i in reshaped_indicator)
        return Counter(chain(g))

    def approximate_slices(self):
        """Identifies the minimal slices in \(S_{11}\) together with their areas and assigns these value to attributes
        `unique_slices` and `unique_slices_areas`."""
        
        print('Slice estimation procedure has started')
        start_time = time.time()

        
        if self.batch_size == None:
            self.batch_size = self.total_nr_samples // 10
        self.total_nr_samples = self.batch_size * (self.total_nr_samples // self.batch_size)

        # rectangle to simulate uniform rvs in space-time is [x, x + ambit_function(0)] x [0,tau]
        low_x, high_x = self.x, self.x + self.ambit_function(0)
        low_t, high_t = max(self.tau + self.decorrelation_time, 0), self.tau
        dict_ = dict()

        #use batches of points to optimise for cache memory and prevent memory overflow 
        for batch in range(self.total_nr_samples // self.batch_size):

            points_x = np.random.uniform(
                low=low_x, high=high_x, size=self.batch_size)
            points_t = np.random.uniform(
                low=low_t, high=high_t, size=self.batch_size)

            # throw away points not contained in A_11 = A + (x,tau):
            # (points_x,points_t) in A_11 if: 0 < points_x - x < phi(points_t - tau)
            # i.e. throw away points contained in ambit sets bottom or left of A_11
            # left of A_11: no such points, since we only simulate points with t coordinate > 0
            # bottom of A_11: must be in A_01; condition: (points_x < phi(points_t - tau)) *  (points_x  > 0)

            indicator_in_A_11 = (
                points_x - self.x < self.ambit_function(points_t - self.tau)) * (points_x - self.x > 0)
            indicator_in_A_01 = (points_x < self.ambit_function(
                points_t - self.tau)) * (points_x > 0)
            indicator_bottom_of_A_11 = indicator_in_A_11 * (~indicator_in_A_01)

            points_x = points_x[indicator_bottom_of_A_11]
            points_t = points_t[indicator_bottom_of_A_11]

            dict_to_add = self.determine_slices_from_points(points_x, points_t)

            for k, v in dict_to_add.items():
                if k in dict_:
                    dict_[k] += v
                else:
                    dict_[k] = v
                    
        # to add more diagnostics
        print('Slice estimation procedure has finished')
        end_time = time.time()
        print(end_time - start_time)
        
        percentage_points_kept = 100 * sum(dict_.values()) / self.total_nr_samples
        print(f"{round(percentage_points_kept,2)}% of points are used in the slice estimation")

        nr_unique_indicators = len(dict_.keys())
        self.unique_slices = np.array(list(dict_.keys())).reshape(
            nr_unique_indicators, self.I_s, self.I_t)
        
        self.unique_slices_areas = np.array(list(dict_.values())) * (high_x-low_x) * \
            (high_t - low_t) / self.total_nr_samples
            
            
    def determine_correction_slices(self,T):
        """Method to be used in the infinite decorrelation time to determine the areas of the
        intersection of the ambit sets at time coordinates \(\\tau,\ldots, k_t \\tau\) with 
        the region of the plane given by \(t < T\). The result is stored in the attribute
        `correction_slices_areas`.
        """
        
        self.correction_slices_areas = [quad(self.ambit_function, a = T - (i+1) * self.tau, b= T - i * self.tau , limit=500)[0]  
                                        for i in range(1,self.k_t)] + [quad(self.ambit_function,a= -np.inf,b=T - self.k_t,limit=500)[0]]
 
        
#    @njit
    def simulate_finite_decorrelation_time(self):
        """implementation of algorithm  [nr to be added] in [paper link]"""
        Y_gaussian = np.zeros((self.nr_simulations,self.k_s + 2 *self.I_s -2,self.k_t + 2 * self.I_t-2))
        Y_jump     = np.zeros((self.nr_simulations,self.k_s + 2 *self.I_s -2,self.k_t + 2 * self.I_t-2))
        for k in range(self.k_s + self.I_s -1):
            for l in range(self.k_t + self.I_t - 1):
                
                gaussian_to_add = np.zeros((self.k_s,self.k_t))
                jump_to_add     = np.zeros((self.k_s,self.k_t))
                
                #simulate S.
                for slice_S,area_S in zip(self.unique_slices,self.unique_slices_areas):
                    tiled_area_S = np.tile(area_S,(self.nr_simulations,1))
                    
                    gaussian_sample_slice = gaussian_part_sampler(self.gaussian_part_params,tiled_area_S)
                    jump_sample_slice     = jump_part_sampler(self.jump_part_params,tiled_area_S,self.jump_part_name)
                    
                    gaussian_to_add += slice_S * gaussian_sample_slice[:,:,np.newaxis]
                    jump_to_add     += slice_S * jump_sample_slice[:,:,np.newaxis]
                    
                Y_gaussian[:,k:k+self.I_s,l:l+self.I_t] +=  gaussian_to_add
                Y_jump[:,k:k+self.I_s,l:l+self.I_t]     +=  jump_to_add
 
        self.gaussian_values =  Y_gaussian[:,self.I_s-1:self.I_s+self.k_s-1,self.I_t-1:self.I_t+self.k_t-1]
        self.jump_values     =  Y_jump[:,self.I_s-1:self.I_s+self.k_s-1,self.I_t-1:self.I_t+self.k_t-1]

#    @njit
    def simulate_infinite_decorrelation_time(self,T):
        """implementation of algorithm  [nr to be added] in [paper link]"""

        assert T/self.tau == int(T/self.tau)
        T_tau = -int(T/self.tau)
        
        Y_gaussian = np.zeros((self.nr_simulations,self.k_s + 2 *self.I_s -2,2 * self.k_t + 2 * T_tau -1))
        Y_jump     = np.zeros((self.nr_simulations,self.k_s + 2 *self.I_s -2,2 * self.k_t + 2 * T_tau -1))
        
        #add correction slices
        self.determine_correction_slices(T)
        correction_slices_matrix   = np.tile(self.correction_slices_areas,(self.nr_simulations,self.k_s,1))
        gaussian_correction_slices = gaussian_part_sampler(self.gaussian_part_params,correction_slices_matrix)
        jump_correction_slices     = jump_part_sampler(self.jump_part_params,correction_slices_matrix,self.jump_part_name)
        
        #gaussian_correction_slices = np.fliplr(np.cumsum(np.fliplr(gaussian_correction_slices),axis=1))
        #jump_correction_slices     = np.fliplr(np.cumsum(np.fliplr(jump_correction_slices),axis=1))
        gaussian_correction_slices = (np.cumsum(gaussian_correction_slices[:,:,::-1],axis=2))[:,:,::-1]
        jump_correction_slices     = (np.cumsum(jump_correction_slices[:,:,::-1],axis=2))[:,:,::-1]

        
        Y_gaussian[:,self.I_s-1:self.I_s - 1+ self.k_s, T_tau:T_tau + self.k_t] +=  gaussian_correction_slices  
        Y_jump[:,self.I_s-1:self.I_s - 1+ self.k_s, T_tau:T_tau + self.k_t]     +=  jump_correction_slices  

        #implementation of algorithm [] from []
        for k in range(self.k_s + self.I_s -1):
            for l in range(self.k_t +  T_tau):

                gaussian_to_add = np.zeros((self.nr_simulations,self.I_s,self.I_t))
                jump_to_add     = np.zeros((self.nr_simulations,self.I_s,self.I_t))
                
                for slice_S,area_S in zip(self.unique_slices,self.unique_slices_areas):
                    tiled_area_S = np.tile(area_S,(self.nr_simulations,1))
                    #simulate S
                    
                    gaussian_sample_slice = gaussian_part_sampler(self.gaussian_part_params,tiled_area_S)
                    jump_sample_slice     = jump_part_sampler(self.jump_part_params,tiled_area_S,self.jump_part_name)
                    
                    gaussian_to_add = slice_S * gaussian_sample_slice[:,:,np.newaxis] 
                    jump_to_add     = slice_S * jump_sample_slice[:,:,np.newaxis] 
                    
                Y_gaussian[:,k:k+self.I_s,l:l+self.I_t] +=  gaussian_to_add
                Y_jump[:,k:k+self.I_s,l:l+self.I_t]     +=  jump_to_add
                    
                    
        self.gaussian_values = Y_gaussian[:,self.I_s-1:self.I_s - 1+ self.k_s,T_tau+1: T_tau+self.k_t+1]
        self.jump_values     = Y_jump[:,self.I_s-1:self.I_s-1+self.k_s,T_tau+1:T_tau+1+self.k_t]
                
    def simulate(self):
        """Simulate the ambit field at time coordinates \(\\tau,\ldots,k_t\\tau\) and space
        coordinates \(x,\ldots,k_s x\). The marginal law of this stationary
        process is given by the independent sum of the Gaussian and jump parts. See [] for an example
        and `helpers.sampler` for the parametrisations used. The simulated values are stored in the
        attribute `values`."""

        # checks
        self.delete_values()
        check_trawl_function(self.ambit_function)
        check_gaussian_params(self.gaussian_part_params)
        check_jump_part_and_params(self.jump_part_name,self.jump_part_params)



        #set I_s
        self.I_s = math.ceil(self.ambit_function(0)/self.x)
        # set I_t
        if self.decorrelation_time > -np.inf:    
            
            # decorrelation_time checks
            assert self.decorrelation_time < 0, 'please check the value of the decorrelation_time'
            assert self.ambit_function(self.decorrelation_time) == 0,\
                            'ambit_function(decorrelation_time) should be 0'
            self.I_t = math.ceil(-self.decorrelation_time/self.tau)
            
            

        elif self.decorrelation_time == -np.inf:
            
            T_tilde = fsolve(lambda t: self.ambit_function(t)-self.x,x0=-1)[0]
            T = self.tau * math.floor(1 + T_tilde/self.tau)
            assert (T/self.tau).is_integer()
            
            self.I_t = int(-T / self.tau) + self.k_t
            
        self.approximate_slices()
        
        if  self.decorrelation_time > -np.inf:           
            self.simulate_finite_decorrelation_time()
            
        elif self.decorrelation_time == -np.inf:
             #self.determine_correction_slices(T)
             self.simulate_infinite_decorrelation_time(T)


        
        self.values = self.gaussian_values + self.jump_values



import numpy  as np

def compute_empirical_covariance_matrix(values,max_lag_space,max_lag_time):
    """
    Args:
      values:
      max_lag_space:
      max_lag_time:
          
    Returns:
      result_cov:
      result_cor
    """
    nr_simulations, nr_rows, nr_columns = values.shape
    result_cov = np.zeros((nr_simulations,max_lag_space+1,max_lag_time+1))
    result_cor = np.zeros((nr_simulations,max_lag_space+1,max_lag_time+1))

    
    for row in range(max_lag_space+1):
        for column in range(max_lag_time+1):
            
            nr_elements = (nr_rows - row)*(nr_columns - column)
            
            sub_matrix_1 = values[:,:nr_rows - row, :nr_columns - column]
            sub_matrix_2 = values[:,row :, column :]
            #assert sub_matrix_1.shape == sub_matrix_2.shape
            
            mean_1  = np.einsum('ijk->i',sub_matrix_1) / nr_elements
            mean_2  = np.einsum('ijk->i',sub_matrix_2) / nr_elements
            
            variance_estimator_1 = np.array([np.var(sub_matrix_1[i,:,:]) for i in range(nr_simulations)])
            variance_estimator_2 = np.array([np.var(sub_matrix_2[i,:,:]) for i in range(nr_simulations)])

            
            sub_matrix_1 = sub_matrix_1 - mean_1[:,np.newaxis,np.newaxis]
            sub_matrix_2 = sub_matrix_2 - mean_2[:,np.newaxis,np.newaxis]
            
            covariances  = np.einsum('ijk,ijk->i',sub_matrix_1,sub_matrix_2) / nr_elements
            
            result_cov[:,row,column] = covariances
            result_cor[:,row,column] = covariances/(variance_estimator_1 * variance_estimator_2)**0.5
            
    return result_cov,result_cor



"""Checks the inputs to the `trawl` and `simple_ambit_field` classes."""

import pandas as pd
import numpy as np

def check1(nr_trawls,nr_simulations):
    #assert (isinstance(tau,float) or isinstance(tau,int)) and tau > 0
    assert isinstance(nr_trawls,int) and nr_trawls >0
    assert isinstance(nr_simulations,int) and nr_simulations >0
    
def check_spatio_temporal_positions(x,tau,k_s,k_t,nr_simulations):
    assert (isinstance(x,float) or isinstance(x,int)) and x > 0
    assert (isinstance(tau,float) or isinstance(tau,int)) and tau > 0
    assert isinstance(k_s,int) and k_s >0
    assert isinstance(k_t,int) and k_t >0
    assert isinstance(nr_simulations,int) and nr_simulations >0


def check_trawl_function(phi):
    """Check if the function is increasing and 0 for strictly positive values"""
    assert callable(phi),'trawl_function is not a function'
    assert phi(0.000001) == 0,'trawl_function does not satisfy trawl_function(t)=0 for t >0'
        
    phi_values = phi(np.linspace(-100,0,10**5))
    assert pd.Series(phi_values).is_monotonic_increasing,'trawl_function is not increasing'   

#distributional checks
        
def check_gaussian_params(gaussian_part_params):
        """Check if the distribution of the jump part is supported and if the parameters 
        of the Gaussian part are numbers"""
        #gaussian part params
        assert isinstance(gaussian_part_params,tuple),'gaussian_part_params is not a tuple'
        assert all(isinstance(i,(int,float)) for i in gaussian_part_params),'parameters of the gaussian part are not numbers'
        
            
def check_jump_part_and_params(jump_part_name,jump_part_params):
        """Check if the distribution of the jump part is supported and if the parameters 
        of the Jump parts are numbers"""
        #jump part params
        assert isinstance(jump_part_params,tuple),'jump_part_params is not a tuple'
        assert all(isinstance(i,(int,float)) for i in jump_part_params),'parameters of the jump part are not numbers'
        
        #jump part_name
        if jump_part_name in ['invgauss','norminvgauss','geninvgauss','nbinom']: #to also add hyperbolic distributions 
            raise ValueError('distribution not yet supported')

        elif jump_part_name not in [None,'gamma','cauchy','poisson']:
            raise ValueError('unknown distribution')
            
def check_grid_params(mesh_size,truncation_grid,times_grid):
    assert  isinstance(mesh_size,(int,float)) and mesh_size >0,'please check mesh size'
    assert  isinstance(truncation_grid,(int,float)) and truncation_grid < 0
            
def check_cpp():
    pass
        


        